"""Domain name asset definition."""
import dataclasses

from ostorlab.assets import asset


@dataclasses.dataclass
@asset.selector('v3.asset.dns.a_record')
class DomainName(asset.Asset):
    """Domain Name target asset per RFC 1034 and 1035."""
    name: str

    def __str__(self):
        return self.name



"""Definition of the emitting protocol exposing the emit method."""
from typing import Any, Dict, Protocol


class EmitProtocol(Protocol):
    """Protocol exposing the emit method."""

    def emit(self, selector: str, data: Dict[str, Any]) -> None:
        """Sends a message to all listening agents on the specified selector.

        Args:
            selector: target selector.
            data: message data to be serialized.
        Raises:
            NonListedMessageSelectorError: when selector is not part of listed out selectors.

        Returns:
            None
        """
        pass



"""Handles calls to the token API.

    Typical usage example:
    public_runner = LoginAPIRunner(username, password, otp_token)
    public_runner.login_user()
"""

import requests
from typing import Dict, Optional

from ostorlab.apis.runners import runner
from ostorlab.apis import request as api_request
from ostorlab.apis import login



TOKEN_ENDPOINT = 'https://api.ostorlab.co/apis/token/'


class LoginAPIRunner(runner.APIRunner):
    """Responsible for the public API calls, and preparing the responses."""
    def __init__(self,
                 username: str = None,
                 password: str = None,
                 otp_token: Optional[str] = None,
                 proxy: str = None,
                 verify: bool = True
                 ):
        """Constructs all the necessary attributes for the object.

        Args:
            username: the username (email) used to login.
            password: the password used to login.
            proxy: The proxy through which a request is made. Defaults to None.
            verify: Whether or not to verify the TLS certificate. Defaults to True.
        """
        super().__init__(proxy, verify)
        self._username = username
        self._password = password
        self._otp_token = otp_token

    @property
    def endpoint(self) -> str:
        """Token API endpoint."""
        return TOKEN_ENDPOINT

    def login_user(self) -> requests.models.Response:
        """Logs in the user.

        Returns:
            The API response.
        """
        login_request = login.UsernamePasswordLoginAPIRequest(self._username,
                                                              self._password,
                                                              self._otp_token)

        return self._sent_request(login_request)


    def execute(self, request: api_request.APIRequest) -> Dict:
        """Executes a request using the Token GraphQL API.

        Args:
            request: The request to be executed

        Raises:
            ResponseError: When the API returns an error

        Returns:
            The API response
        """
        response = self._sent_request(request)
        if response.status_code != 200:
            raise runner.ResponseError(
                f'Response status code is {response.status_code}: {response.content}')
        else:
            return response.json()



"""Runtime agent module gives access to methods to create the specific agent configurations & run the agent.

Usage
    agent_runtime = AgentRuntime(agent_settings, runtime_name, docker_client, mq_service)
    agent_service = agent_runtime.create_agent_service(network_name, extra_configs)
"""
from typing import List, Optional
import docker
from docker.types import services as docker_types_services
import io

from ostorlab.runtimes.local import runtime
from ostorlab.runtimes import definitions
from ostorlab.agent import definitions as agent_definitions
from ostorlab.runtimes.local.services import mq
from ostorlab import configuration_manager

MOUNT_VARIABLES = {
    '$CONFIG_HOME': str(configuration_manager.OSTORLAB_PRIVATE_DIR)
}

class AgentRuntime:
    """Class to consolidate the agent settings and agent default definition, and create the agent service."""

    def __init__(self,
                 agent_settings: definitions.AgentSettings,
                 runtime_name: str,
                 docker_client: docker.DockerClient,
                 mq_service: mq.LocalRabbitMQ) -> None:
        """Constructs all the necessary attributes for the object.

        Args:
            agent: an agent definition containing all the settings of how agent should run and what arguments to pass.
            runtime_name: local runtime instance name.
            docker_client: docker client.
        """
        self._docker_client = docker_client
        self.agent = agent_settings
        self.image_name = agent_settings.container_image.split(':', maxsplit=1)[0]
        self.runtime_name = runtime_name
        self.mq_service = mq_service
        self.update_agent_settings()

    def create_settings_config(self) -> docker.types.ConfigReference:
        """Create a docker configuration of the  agent settings.

        Returns:
            docker ConfigReference of the settings configuration
        """
        agent_instance_settings_proto = self.agent.to_raw_proto()
        config_name = f'config_settings_{self.image_name}_{self.runtime_name}'
        docker_config = self._docker_client.configs.create(name=config_name,
                                                           labels={'ostorlab.universe': self.runtime_name},
                                                           data=agent_instance_settings_proto)
        return docker.types.ConfigReference(config_id=docker_config.id,
                                            config_name=config_name,
                                            filename='/tmp/settings.binproto')

    def create_definition_config(self) -> docker.types.ConfigReference:
        """Create a docker configuration of the  agent definition.

        Returns:
            docker configuration reference of the agent defintion configuration.
        """
        agent_definition = self._docker_client.images.get(self.agent.container_image).labels.get('agent_definition')
        config_name = f'config_definition_{self.image_name}__{self.runtime_name}'
        docker_config = self._docker_client.configs.create(name=config_name,
                                                           labels={'ostorlab.universe': self.runtime_name},
                                                           data=str.encode(agent_definition))
        return docker.types.ConfigReference(config_id=docker_config.id,
                                            config_name=config_name,
                                            filename='/tmp/ostorlab.yaml')

    def create_agent_definition_from_label(self) -> agent_definitions.AgentDefinition:
        """Read the agent yaml definition from the docker image labels.

        Returns:
            the agent defintion.
        """
        docker_image = self._docker_client.images.get(self.agent.container_image)
        yaml_definition_string = docker_image.labels.get('agent_definition')
        with io.StringIO(yaml_definition_string) as file:
            agent_definition = agent_definitions.AgentDefinition.from_yaml(file)
            return agent_definition

    def update_agent_settings(self) -> None:
        """Update agent settings with values from the local runtime."""
        self.agent.bus_url = self.mq_service.url
        self.agent.bus_exchange_topic = f'ostorlab_topic_{self.runtime_name}'
        self.agent.bus_management_url = self.mq_service.management_url
        self.agent.bus_vhost = self.mq_service.vhost
        self.agent.healthcheck_host = runtime.HEALTHCHECK_HOST
        self.agent.healthcheck_port = runtime.HEALTHCHECK_PORT

    def create_docker_healthchek(self) -> docker.types.Healthcheck:
        """Create a docker healthcheck configuration for for the agent service.

        Returns:
            docker healthcheck configuration.
        """
        # wait 2s and check max 5 times with 0.5s between each check.
        healthcheck = docker.types.Healthcheck(test=['CMD', 'ostorlab', 'agent', 'healthcheck'],
                                               retries=runtime.HEALTHCHECK_RETRIES,
                                               timeout=runtime.HEALTHCHECK_TIMEOUT,
                                               start_period=runtime.HEALTHCHECK_START_PERIOD,
                                               interval=runtime.HEALTHCHECK_INTERVAL, )
        return healthcheck

    def replace_variable_mounts(self, mounts: List[str]):
        """Replace path variables for the container mounts

        Args:
            mounts: List of src:dst paths to mount
        """

        replaced_mounts= []
        for mount in mounts:
            for mount_variable, mount_value in MOUNT_VARIABLES.items():
                mount = mount.replace(mount_variable, mount_value)
            replaced_mounts.append(mount)
        return replaced_mounts


    def create_agent_service(self,
                            network_name: str,
                            extra_configs: Optional[List[docker.types.ConfigReference]] = None
                            ) -> docker.models.services.Service:
        """Create the agent service.

        Args:
            network_name: network name to attach the service to.
            extra_configs: list of docker ConfigReferences that will be exposed to the service.

        Returns:
            the agent docker service.
        """
        agent_definition = self.create_agent_definition_from_label()
        self.agent.open_ports = self.agent.open_ports or agent_definition.open_ports
        if self.agent.open_ports:
            endpoint_spec = docker_types_services.EndpointSpec(
                ports={p.destination_port: p.source_port for p in self.agent.open_ports})
        else:
            endpoint_spec = docker_types_services.EndpointSpec(mode='dnsrr')

        extra_configs.append(self.create_settings_config())
        extra_configs.append(self.create_definition_config())

        mounts = self.agent.mounts or agent_definition.mounts
        mounts = self.replace_variable_mounts(mounts)
        constraints = self.agent.constraints or agent_definition.constraints
        mem_limit = self.agent.mem_limit or agent_definition.mem_limit
        restart_policy = self.agent.restart_policy or agent_definition.restart_policy

        service_name = self.agent.container_image.replace(':', '_').replace('.','') + '_' + self.runtime_name
        agent_service = self._docker_client.services.create(
            image=self.agent.container_image,
            networks=[network_name],
            env=[
                f'UNIVERSE={self.runtime_name}',
            ],
            name=service_name,
            restart_policy=docker_types_services.RestartPolicy(condition=restart_policy),
            mounts=mounts,
            healthcheck=self.create_docker_healthchek(),
            labels={'ostorlab.universe': self.runtime_name},
            configs=extra_configs,
            constraints=constraints,
            endpoint_spec=endpoint_spec,
            resources=docker_types_services.Resources(mem_limit=mem_limit))

        return agent_service



"""Agent Build commands."""
import io
import logging

import click
import docker

from ostorlab.agent.schema import loader
from ostorlab.agent.schema import validator
from ostorlab.cli import console as cli_console
from ostorlab.cli.agent import agent
from ostorlab.cli import docker_requirements_checker

console = cli_console.Console()

logger = logging.getLogger(__name__)


@agent.command()
@click.option('--file', '-f', type=click.File('rb'), help='Path to Agent yaml definition.', required=True)
@click.option('--organization', '-o', help='Organization name.', required=False, default='')
def build(file: io.FileIO, organization: str = '') -> None:
    """CLI command to build the agent container from a definition.yaml file.
    Usage : Ostorlab agent build -f path/to/definition.yaml -org organization_name
    """

    if not docker_requirements_checker.is_docker_installed():
        console.error('Docker is not installed.')
        raise click.exceptions.Exit(2)
    elif not docker_requirements_checker.is_user_permitted():
        console.error('User does not have permissions to run docker.')
        raise click.exceptions.Exit(2)
    else:
        try:
            agent_def = loader.load_agent_yaml(file)
            file.seek(0)
            dockerfile_path = agent_def['docker_file_path']
            docker_build_root = agent_def['docker_build_root']
            agent_name = agent_def['name']
            agent_version = agent_def.get('version', '0.0.0')
            container_name = f'agent_{organization}_{agent_name}:v{agent_version}'
            docker_sdk_client = docker.from_env()

            try:
                docker_sdk_client.images.get(container_name)
                console.info(f'{container_name} already exist.')
                raise click.exceptions.Exit(0)
            except docker.errors.ImageNotFound:
                console.info(
                    f'Building agent [bold red]{agent_name}[/] dockerfile [bold red]{dockerfile_path}[/]'
                    f' at root [bold red]{docker_build_root}[/].')
                with console.status(f'Building [bold red]{container_name}[/]'):
                    docker_sdk_client.images.build(path=docker_build_root,
                                                   dockerfile=dockerfile_path,
                                                   tag=container_name,
                                                   labels={'agent_definition':file.read().decode('utf-8')})
                console.success(f'Agent {agent_name} built, container [bold red]{container_name}[/] created.')
        except validator.SchemaError:
            console.error(
                'Schema is invalid, this should not happen, please report an issue at '
                'https://github.com/Ostorlab/ostorlab/issues.')
        except validator.ValidationError:
            console.error('Definition file does not conform to the provided specification.')



"""Module contains all the supported assets that we can pass to the scan command."""
from ostorlab.cli.scan.run.assets import android_apk
from ostorlab.cli.scan.run.assets import android_aab
from ostorlab.cli.scan.run.assets import ios_ipa
from ostorlab.cli.scan.run.assets import file
from ostorlab.cli.scan.run.assets import ip



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: web.proto

import sys

_b = sys.version_info[0] < 3 and (lambda x: x) or (lambda x: x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()

DESCRIPTOR = _descriptor.FileDescriptor(
    name='web.proto',
    package='',
    serialized_pb=_b(
        '\n\tweb.proto\"E\n\x03web\x12\x0f\n\x07scan_id\x18\x01 \x02(\x05\x12\x16\n\x04urls\x18\x02 \x03(\x0b\x32\x08.web.url\x1a\x15\n\x03url\x12\x0e\n\x06\x64omain\x18\x01 \x02(\t')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

_WEB_URL = _descriptor.Descriptor(
    name='url',
    full_name='web.url',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='domain', full_name='web.url.domain', index=0,
            number=1, type=9, cpp_type=9, label=2,
            has_default_value=False, default_value=_b("").decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    options=None,
    is_extendable=False,
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=61,
    serialized_end=82,
)

_WEB = _descriptor.Descriptor(
    name='web',
    full_name='web',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='scan_id', full_name='web.scan_id', index=0,
            number=1, type=5, cpp_type=1, label=2,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='urls', full_name='web.urls', index=1,
            number=2, type=11, cpp_type=10, label=3,
            has_default_value=False, default_value=[],
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
    ],
    extensions=[
    ],
    nested_types=[_WEB_URL, ],
    enum_types=[
    ],
    options=None,
    is_extendable=False,
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=13,
    serialized_end=82,
)

_WEB_URL.containing_type = _WEB
_WEB.fields_by_name['urls'].message_type = _WEB_URL
DESCRIPTOR.message_types_by_name['web'] = _WEB

web = _reflection.GeneratedProtocolMessageType('web', (_message.Message,), dict(

    url=_reflection.GeneratedProtocolMessageType('url', (_message.Message,), dict(
        DESCRIPTOR=_WEB_URL,
        __module__='web_pb2'
        # @@protoc_insertion_point(class_scope:web.url)
    ))
    ,
    DESCRIPTOR=_WEB,
    __module__='web_pb2'
    # @@protoc_insertion_point(class_scope:web)
))
_sym_db.RegisterMessage(web)
_sym_db.RegisterMessage(web.url)

# @@protoc_insertion_point(module_scope)



"""Vulnz Describe command."""
import logging
import click

import rich
from rich import markdown
from rich import panel

from ostorlab.cli import console as cli_console
from ostorlab.cli.vulnz import vulnz
from ostorlab.runtimes.local.models import models
from ostorlab.utils import styles


console = cli_console.Console()

logger = logging.getLogger(__name__)


@vulnz.command(name='describe')
@click.option('--vuln-id', '-v', 'vuln_id', help='Id of the vulnerability.', required=True)
def describe_cli(vuln_id: int) -> None:
    """CLI command to describe a vulnerability."""
    database = models.Database()
    session = database.session
    vulnerability = session.query(models.Vulnerability).get(vuln_id)
    console.success('Vulnerabilities retrieved successfully.')
    vulnz_list = [
        {'id': str(vulnerability.id),
            'risk_rating': styles.style_risk(vulnerability.risk_rating.value.upper()),
            'cvss_v3_vector': vulnerability.cvss_v3_vector,
            'title': vulnerability.title,
            'short_description': markdown.Markdown(vulnerability.short_description),
         }
    ]

    columns = {
        'Id': 'id',
        'Title': 'title',
        'Risk rating': 'risk_rating',
        'CVSS V3 Vector': 'cvss_v3_vector',
        'Short Description': 'short_description',
    }
    title = f'Describing vulnerability {vuln_id}'
    console.table(columns=columns, data=vulnz_list, title=title)
    rich.print(panel.Panel(markdown.Markdown(vulnerability.description), title='Description'))
    rich.print(panel.Panel(markdown.Markdown(vulnerability.recommendation), title='Recommendation'))
    rich.print(panel.Panel(markdown.Markdown(vulnerability.technical_detail), title='Technical details'))




"""mock agent implements the required methods to test the agent's behavior without using external components."""
import pytest

from typing import List

from ostorlab.agent import message as msg


@pytest.fixture(scope='function')
def agent_mock(mocker):
    """This fixtures patches all the Agent components and returns the list of messages emitted """
    emitted_messages: List[msg.Message] = []
    mocker.patch('ostorlab.agent.mixins.agent_mq_mixin.AgentMQMixin.mq_init', return_value=None)
    mocker.patch('ostorlab.agent.mixins.agent_mq_mixin.AgentMQMixin.mq_run', return_value=None)
    mocker.patch('ostorlab.agent.mixins.agent_mq_mixin.AgentMQMixin.mq_close', return_value=None)
    mocker.patch('ostorlab.agent.agent.AgentMixin._is_mq_healthy', return_value=True)
    mocker.patch(
        'ostorlab.agent.mixins.agent_healthcheck_mixin.AgentHealthcheckMixin.start_healthcheck',
        return_value=None
    )
    mocker.patch(
        'ostorlab.agent.mixins.agent_healthcheck_mixin.AgentHealthcheckMixin.add_healthcheck',
        return_value=None
    )
    mocker.patch(
        'ostorlab.agent.mixins.agent_healthcheck_mixin.AgentHealthcheckMixin.__init__',
        return_value=None
    )

    def mq_send_message(key, message):
        # we need to remove the last part of the key f'{selector}.{uuid.uuid1()}'
        agent_message = msg.Message.from_raw('.'.join(key.split('.')[:-1]), message)
        emitted_messages.append(agent_message)

    mocker.patch('ostorlab.agent.mixins.agent_mq_mixin.AgentMQMixin.mq_send_message', side_effect=mq_send_message)
    yield emitted_messages
    emitted_messages = []







"""Asset of type file.
This module takes care of preparing a generic file of type before injecting it to the runtime instance."""
import io
import click
from ostorlab.assets import file as file_asset
from ostorlab.cli.scan.run import run


@run.run.command(name='file')
@click.option('--file', type=click.File(mode='rb'), help='Path to file.', required=True)
@click.pass_context
def file_cli(ctx: click.core.Context, file: io.FileIO) -> None:
    """Run scan for file asset."""
    runtime = ctx.obj['runtime']
    asset = file_asset.File(content=file.read(), path=file.name)
    runtime.scan(title=ctx.obj['title'], agent_group_definition=ctx.obj['agent_group_definition'], asset=asset)



"""Vulnz module that handles listing and describing a vulnerability."""
from ostorlab.cli.rootcli import rootcli


@rootcli.group()
def vulnz() -> None:
    """You can use vulnz to list and describe vulnerabilities.\n"""
    pass



"""Vulnz List command."""
import logging
import click

from rich import markdown

from ostorlab.cli import console as cli_console
from ostorlab.cli.vulnz import vulnz
from ostorlab.runtimes.local.models import models
from ostorlab.utils import styles

console = cli_console.Console()

logger = logging.getLogger(__name__)


@vulnz.command(name='list')
@click.option('--scan-id', '-s', 'scan_id', help='Id of the scan.', required=True)
def list_cli(scan_id: int) -> None:
    """CLI command to list vulnerabilities for a scan."""
    database = models.Database()
    session = database.session
    vulnerabilities = session.query(models.Vulnerability).filter_by(scan_id=scan_id).\
        order_by(models.Vulnerability.title).all()
    console.success('Vulnerabilities listed successfully.')
    vulnz_list = []
    for vulnerability in vulnerabilities:
        vulnz_list.append({
            'id': str(vulnerability.id),
            'risk_rating': styles.style_risk(vulnerability.risk_rating.value.upper()),
            'cvss_v3_vector': vulnerability.cvss_v3_vector,
            'title': vulnerability.title,
            'short_description': markdown.Markdown(vulnerability.short_description),
        })

    columns = {
        'Id': 'id',
        'Title': 'title',
        'Risk rating': 'risk_rating',
        'CVSS V3 Vector': 'cvss_v3_vector',
        'Short Description': 'short_description',
    }
    title = f'Found {len(vulnz_list)} vulnerabilities'
    console.table(columns=columns, data=vulnz_list, title=title)



"""Serializer handles matching of selector to the proper protobuf message definition."""
import importlib
import logging
import os
import pathlib
import re
from typing import Dict, Optional, List, Any

from google.protobuf import json_format

from ostorlab import exceptions

PROTO_CLASS_NAME = 'Message'

# When the package is built, the proto files are not shipped, only the compiled version, which have the
# form: logs_pb2.cpython-38.pyc
MESSAGE_PROTO = '_pb2'
MESSAGE_CODE_PATH = pathlib.Path(__file__).parent / 'proto'

logger = logging.getLogger(__name__)

class SerializationError(exceptions.OstorlabError):
    """Base serialization Error."""


class TooManyMatchingPackageNamesError(SerializationError):
    """There are over 2 matching package names."""


class NoMatchingPackageNameError(SerializationError):
    """There are no matching package name."""


def _find_package_name(selector: str) -> Optional[str]:
    """Finds matching package name to selector."""
    files = _list_message_proto_files()
    regex_pattern = _selector_to_package_regex(selector)
    logger.debug('searching protos with pattern: %s', regex_pattern)
    pattern = re.compile(regex_pattern)
    matching = [pattern.match(f) for f in files]
    filtered_matching = [m.group(0) for m in matching if m is not None]
    if len(filtered_matching) > 1:
        raise TooManyMatchingPackageNamesError(f'Found {",".join(filtered_matching)}')
    elif len(filtered_matching) == 0:
        raise NoMatchingPackageNameError()
    else:
        # Remove the path to the current package.
        matching_package = re.sub(r'^.*/ostorlab/', 'ostorlab/', filtered_matching[0])
        # Replace / with .
        matching_package = matching_package.replace('/', '.')
        # Point to the compiled protobufs.
        return re.sub(r'_pb2\..*$', '_pb2', re.sub(r'^\.code\.', '', matching_package))


def _list_message_proto_files() -> List[str]:
    """List all the proto files."""
    files = []
    # r=root, d=directories, f = files
    for r, d, f in os.walk(MESSAGE_CODE_PATH):
        for file in f:
            if MESSAGE_PROTO in file:
                files.append(os.path.join(r, file))
        del d
    return files


def _selector_to_package_regex(subject):
    """Maps selector to package matching regular expression."""
    splitted = subject.split('.')
    return '.*/message/proto/' + '/'.join([f'(_[_a-zA-Z0-9]+|{s})' for s in splitted]) + r'.[_a-zA-Z0-9]+\_pb2\.py'


def serialize(selector: str, values: Dict[str, Any]):
    """Serializes a Request message using the proper format defined using the seelctor value.
    If the subject is a.b.c. The corresponding proto is located at message/a/b/c/xxx.proto.

    Args:
        selector: Message selector, must specify the version in use.
        values: Dict representation of the message to serialize.

    Returns:
        Proto serialized message.
    """
    try:
        return _serialize(selector, PROTO_CLASS_NAME, values)
    except json_format.Error as e:
        raise SerializationError('Error serializing message') from e


def _serialize(selector: str, class_name: str, values: Dict[str, Any]):
    """Serializes message using the selector and defined class name."""
    package_name = _find_package_name(selector)
    class_object = getattr(importlib.import_module(package_name), class_name)
    proto_message = class_object()
    _parse_dict(values, proto_message)
    return proto_message


def _parse_list(values: Any, message) -> None:
    """Parse list to protobuf message."""
    if len(values) > 0 and isinstance(values[0], dict):  # value needs to be further parsed
        for v in values:
            cmd = message.add()
            _parse_dict(v, cmd)
    else:  # value can be set
        message.extend(values)


def _parse_dict(values: Any, message) -> None:
    """Parse dict to protobuf message."""
    for k, v in values.items():
        if isinstance(v, dict):  # value needs to be further parsed
            _parse_dict(v, getattr(message, k))
        elif isinstance(v, list):
            _parse_list(v, getattr(message, k))
        else:
            try:
                # if is of type ENUM.
                if message.DESCRIPTOR.fields_by_name[k].type == 14:
                    # For type enum, we introspect the type to get enum type, and then get the string mapping to int.
                    enum_v = message.DESCRIPTOR.fields_by_name[k].enum_type.values_by_name[v].number
                    setattr(message, k, enum_v)
                elif v is None:
                    # Optional values don't need to be set.
                    pass
                else:
                    setattr(message, k, v)
            except AttributeError as e:
                raise SerializationError(f'invalid attribute {k}') from e
            except KeyError as e:
                raise SerializationError(f'invalid attribute {k}') from e


def deserialize(selector: str, serialized: bytes):
    """Deserializes a Request message using the proper format defined using the selector value.

    Args:
        selector: Message selector, must specify the version in use.
        serialized: Raw message to deserialize.

    Returns:
        Dict
    """
    return _deserialize(selector, PROTO_CLASS_NAME, serialized)


def _deserialize(selector: str, class_name: str, serialized: bytes):
    """Deserializes message using the selector and defined class name."""
    package_name = _find_package_name(selector)
    class_object = getattr(importlib.import_module(package_name), class_name)
    proto_message = class_object()
    proto_message.ParseFromString(serialized)
    return proto_message



"""Module for the command run inside the group scan.
This module takes care of preparing the selected runtime and the lists of provided agents, before starting a scan.
Example of usage:
    - ostorlab scan run --agents=agent1,agent2 --title=test_scan [asset] [options]."""
import io
from typing import List

import click

from ostorlab.cli import install_agent
from ostorlab.cli.scan import scan
from ostorlab.runtimes import definitions
from ostorlab.runtimes import runtime


@scan.group()
@click.option('--agents',
              multiple=True,
              help='List of agents keys. to use in the scan. ',
              required=False)
@click.option('--title', '-t', help='Scan title.')
@click.option('--agent-group-definition', '-g', type=click.File('r'),
              help='Path to agents group definition file (yaml).',
              required=False)
@click.option('--install', '-i', help='Install missing agents', is_flag=True, required=False)
@click.pass_context
def run(ctx: click.core.Context, agents: List[str], agent_group_definition: io.FileIO,
        title: str, install: bool) -> None:
    """Start a new scan on a specific asset.\n
    Example:\n
        - ostorlab scan run --agents=agent/ostorlab/nmap,agent/google/tsunami --title=test_scan ip 8.8.8.8
    """
    if agents:
        agents_settings: List[definitions.AgentSettings] = []
        for agent_key in agents:
            agents_settings.append(
                definitions.AgentSettings(key=agent_key))
            if install:
                install_agent.install(agent_key)
        agent_group = definitions.AgentGroupDefinition(agents=agents_settings)
    elif agent_group_definition:
        agent_group = definitions.AgentGroupDefinition.from_yaml(agent_group_definition)
    else:
        raise click.ClickException('Missing agent list or agent group definition.')

    runtime_instance: runtime.Runtime = ctx.obj['runtime']
    if runtime_instance.can_run(agent_group_definition=agent_group):
        ctx.obj['agent_group_definition'] = agent_group
        ctx.obj['title'] = title

        if install:
            # Trigger both the runtime installation routine and install all the provided agents.
            runtime_instance.install()
            for agent in agent_group.agents:
                install_agent.install(agent.key)
    else:
        raise click.ClickException('The runtime does not support the provided agent list or group definition.')



"""Agent List command."""
import logging
import re
from typing import Optional

import click
import docker

from ostorlab.cli import console as cli_console
from ostorlab.cli.agent import agent as agent_cli

console = cli_console.Console()

logger = logging.getLogger(__name__)


@agent_cli.command()
@click.argument('agent', required=True)
@click.option('--agent-version-regex', '-r', help='Agent version matching regular expression.', required=False)
def delete(agent: str, agent_version_regex: Optional[str] = None) -> None:
    """CLI command to delete installed agents."""
    docker_client = docker.from_env()
    images = docker_client.images.list()
    agent_container_name = agent.replace('/', '_')
    deleted = False
    for im in images:
        for t in im.tags:
            # The agents when they are installed, produce 2 images, one tagged from the agent name and one with the
            # remote name. Hence we need to delete both.
            if t.split(':')[0] in (agent_container_name, f'ostorlab.store/agents/{agent_container_name}'):
                agent_container_version = t.split(':')[1]
                if agent_version_regex is None or re.match(agent_version_regex, agent_container_version):
                    console.info(f'deleting container container [bold red]{t}[/]')
                    docker_client.images.remove(t, force=True)
                    console.success(f'container image [bold red]{t}[/] deleted successfully')
                    deleted = True

    if deleted is False:
        console.error(f'No agent matching [bold white]{agent}[/] was found')



"""Agent search command."""
import logging
from typing import Dict
import click

from ostorlab.cli import console as cli_console
from ostorlab.cli.agent import agent
from ostorlab.apis.runners import runner as base_runner
from ostorlab import configuration_manager
from ostorlab.apis.runners import public_runner, authenticated_runner
from ostorlab.apis import agent_search as agent_search_api

console = cli_console.Console()

logger = logging.getLogger(__name__)


@agent.command(name='search')
@click.option('--keyword', '-k', help='Keyword to use for the search.', required=True)
def search_cli(keyword: str) -> None:
    """CLI command to search for agents in the Agents store."""
    result_agents = search_agents(keyword).get('agents')
    console.success('Search agents done.')
    agents = []
    version = ''
    description = ''
    in_selectors = ''
    out_selectors = ''

    if result_agents is not None:
        for result_agent in result_agents:
            versions = result_agent['versions'].get('versions')
            if len(versions) > 0:
                version = versions[0].get('version')
                description = versions[0].get('description')
                in_selectors = ', '.join(versions[0].get('inSelectors'))
                out_selectors = ',' .join(versions[0].get('outSelectors'))
            else:
                version = '-'
                description = '-'
                in_selectors = '-'
                out_selectors = '-'

            agents.append({
                'key': result_agent['key'],
                'version': version,
                'description': description,
                'in_selectors': in_selectors,
                'out_selectors': out_selectors
            })

    columns = {
        'Agent key': 'key',
        'Version': 'version',
        'Description': 'description',
        'In selectors': 'in_selectors',
        'Out selectors': 'out_selectors',
    }
    title = f'Found {len(agents)} Agents'
    console.table(columns=columns, data=agents, title=title)


def search_agents(search: str) -> Dict:
    """Sends an API request with the agent key, and retrieve the agent information.

    Args:
        search: the keyword to search for

    Returns:
        List of agent matching the search.

    Raises:
        click Exit exception with satus code 2 when API response is invalid.
    """
    config_manager = configuration_manager.ConfigurationManager()

    if config_manager.is_authenticated:
        runner = authenticated_runner.AuthenticatedAPIRunner()
    else:
        runner = public_runner.PublicAPIRunner()

    try:
        response = runner.execute(agent_search_api.AgentSearchAPIRequest(search))
    except base_runner.ResponseError as e:
        console.error('Requested resource not found.')
        raise click.exceptions.Exit(2) from e

    if 'errors' in response:
        error_message = f'The provided search : {search} does not correspond to any agent.'
        console.error(error_message)
        raise click.exceptions.Exit(2)
    else:
        agent_details = response['data']['agents']
        return agent_details



"""Asset of type .IPA (IOS App Store Package).
This module takes care of preparing a file of type .IPA before injecting it to the runtime instance.
"""

import click
from ostorlab.assets import ios_ipa as ios_ipa_asset
from ostorlab.cli.scan.run import run


@run.run.command()
@click.option('--file', type=click.File(mode='rb'), help='Path to .IPA file.', required=True)
@click.pass_context
def ios_ipa(ctx, file):
    """Run scan for .IPA package file."""
    runtime = ctx.obj['runtime']
    asset = ios_ipa_asset.IOSIpa(content=file.read(), path=file.name)
    runtime.scan(title=ctx.obj['title'], agent_group_definition=ctx.obj['agent_group_definition'], asset=asset)



"""Validation of Yaml configuration files against json schema files.

    Typical usage example:
    validator = Validator(json_schema_file_object)
    validator.validate(yaml_file_object)
"""
import io
import json

import jsonschema
import ruamel.yaml

from ostorlab import exceptions


class Error(exceptions.OstorlabError):
    """Base Exception
    """


class ValidationError(Error):
    """Wrapper Exception for the ValidationError produced by jsonschema's validate method."""


class SchemaError(Error):
    """Wrapper Exception for the SchemaError produced by jsonschema's validate method."""


class Validator:
    """Creates validator that checks yaml files with a json schema."""

    def __init__(self, json_schema_file_object: io.FileIO):
        """Inits Validator class.

        Args:
            json_schema_file_object: A file object of the Json schema file.

        Raises:
            SchemaError: When the Json schema file in itself is not valid.
        """
        self._json_schema = json.load(json_schema_file_object)
        try:
            jsonschema.Draft202012Validator.check_schema(self._json_schema)
        except jsonschema.exceptions.SchemaError as e:
            raise SchemaError('Schema is invalid.') from e

    def validate(self, yaml_file_object):
        """ Validates a yaml file against a json schema .

        Args:
            yaml_file_object: A file object of the yaml configuration file we want to validate.
        Raises:
            ValidationError if the validation did not pass well.
            SchemaError if the Schema is not valid.
        """
        yaml = ruamel.yaml.YAML(typ='safe')
        yaml_data = yaml.load(yaml_file_object)

        try:
            jsonschema.validate(instance=yaml_data, schema=self._json_schema)
        except jsonschema.exceptions.ValidationError as e:
            raise ValidationError('Validation did not pass well.') from e
        except jsonschema.exceptions.SchemaError as e:
            raise SchemaError('Schema is invalid.') from e



"""Remote runtime runs on Ostorlab cloud.

The remote runtime provides capabilities identical to local runtime with extra features, like data persistence,
improved data visualization, automated scaling for improved performance, agent improved data warehouse for improved
detection and several other improvements.
"""
from typing import List

from ostorlab.apis import scan_list
from ostorlab.apis import scan_stop
from ostorlab.apis.runners import authenticated_runner
from ostorlab.apis.runners import runner
from ostorlab.assets import asset as base_asset
from ostorlab.cli import console as cli_console
from ostorlab.runtimes import definitions
from ostorlab.runtimes import runtime

console = cli_console.Console()


class RemoteRuntime(runtime.Runtime):
    """Remote runtime instance running on Ostorlab cloud.

    Remote runtime communicates over a GraphQL API."""

    def can_run(self, agent_group_definition: definitions.AgentGroupDefinition) -> bool:
        """Checks if the runtime is capable of running the provided agent run definition.

        Args:
            agent_group_definition: The agent run definition from a set of agents and agent groups.

        Returns:
            True if can run, false otherwise.
        """
        pass

    def scan(self, agent_group_definition: definitions.AgentGroupDefinition, asset: base_asset.Asset) -> None:
        """Triggers a scan using the provided agent run definition and asset target.

        Args:
            agent_group_definition: The agent run definition from a set of agents and agent groups.
            asset: The scan target asset.

        Returns:
            None
        """
        pass

    def stop(self, scan_id: int) -> None:
        """Stops a scan.

        Args:
            scan_id: The id of the scan to stop.
        """
        try:
            api_runner = authenticated_runner.AuthenticatedAPIRunner()
            response = api_runner.execute(scan_stop.ScanStopAPIRequest(scan_id))
            if response.get('errors') is not None:
                console.error(f'Scan with id {scan_id} not found')
            else:
                console.success('Scan stopped successfully')
        except runner.Error:
            console.error('Could not stop scan.')

    def list(self, page: int = 1, number_elements: int = 10) -> List[runtime.Scan]:
        """Lists scans managed by runtime.

        Args:
            page: Page number for list pagination (default 1).
            number_elements: count of elements to show in the listed page (default 10).

        Returns:
            List of scan objects.
        """
        try:
            api_runner = authenticated_runner.AuthenticatedAPIRunner()
            response = api_runner.execute(scan_list.ScansListAPIRequest(page, number_elements))
            scans = response['data']['scans']['scans']

            return [
                runtime.Scan(
                    id=scan['id'],
                    asset=scan['asset'],
                    created_time=scan['createdTime'],
                    progress=scan['progress'],
                ) for scan in scans
            ]

        except runner.Error:
            console.error('Could not fetch scans.')

    def install(self) -> None:
        """No installation action.

        Returns:
            None
        """
        pass



"""Version utility class to make easy to compare semantic version strings."""
import semver


class Version:
    """Version class that only accepts semantic and provides python comparison API."""

    def __init__(self, version: str) -> None:
        """Init version."""
        self._version = version
        self._semver: semver.VersionInfo = semver.VersionInfo.parse(version)

    def __repr__(self):
        """Version string representation."""
        return f'<Version {self._version}>'

    def __str__(self):
        """Version string"""
        return self._version

    def __lt__(self, other):
        """Pythonic comparison API."""
        if isinstance(other, Version):
            return self._semver.compare(other._version) < 0
        else:
            raise ValueError()

    def __le__(self, other):
        """Pythonic comparison API."""
        if isinstance(other, Version):
            return self._semver.compare(other._version) <= 0
        else:
            raise ValueError()

    def __gt__(self, other):
        """Pythonic comparison API."""
        if isinstance(other, Version):
            return self._semver.compare(other._version) > 0
        else:
            raise ValueError()

    def __ge__(self, other):
        """Pythonic comparison API."""
        if isinstance(other, Version):
            return self._semver.compare(other._version) >= 0
        else:
            raise ValueError()

    def __eq__(self, other):
        """Pythonic comparison API."""
        if isinstance(other, Version):
            return self._version == other._version
        else:
            raise ValueError()



"""Check if requirements for running docker are satisfied."""

import docker

def is_docker_installed() -> bool:
    """Checks if docker is installed

    Returns:
        True if docker is installed, else False
    """
    try:
        _ = docker.from_env()
    except docker.errors.DockerException as e:
        if 'ConnectionRefusedError' in str(e):
            return False
    return True

def is_user_permitted() -> bool:
    """Check if the user got permissions to run docker.

    Returns:
        True if user has permission to run docker, else False
    """
    try:
        _ = docker.from_env()
    except docker.errors.DockerException as e:
        if 'PermissionError' in str(e):
            return False
    return True

def is_swarm_initialized() -> bool:
    """Checks if docker swarm is initialized.

    Returns:
        True if docker swarm is initialized, else False
    """
    if is_user_permitted():
        docker_client = docker.from_env()
        if docker_client.swarm.id is None:
            return False
        else:
            return True

def init_swarm() -> None:
    """Initialize docker swarm"""
    if is_user_permitted():
        docker_client = docker.from_env()
        docker_client.swarm.init()



"""Vulnz dump command."""
import logging
import click

from sqlalchemy import case

from ostorlab.cli import console as cli_console
from ostorlab.cli.vulnz import vulnz
from ostorlab.cli import dumpers
from ostorlab.runtimes.local.models import models


console = cli_console.Console()

logger = logging.getLogger(__name__)


@vulnz.command(name='dump')
@click.option('--scan-id', '-s', 'scan_id', help='Id of the scan.', required=True)
@click.option('--output', '-o', help='Output file path', required=False, default='./output.json')
@click.option('--format', '-f', 'output_format', help='Output format',
              required=False, type=click.Choice(['json', 'csv']), default='json')
def dump(scan_id: int, output: str, output_format: str) -> None:
    """Dump found vulnerabilities of a scan in a specific format."""
    database = models.Database()
    session = database.session
    severities = {'HIGH': 0, 'MEDIUM': 1, 'LOW': 2, 'POTENTIALLY': 3,
                  'HARDENING': 4, 'SECURE': 5, 'IMPORTANT': 6, 'INFO': 7}
    severity_sort_logic = case(value=models.Vulnerability.risk_rating, whens=severities).label('severity')
    vulnerabilities = session.query(models.Vulnerability).filter_by(scan_id=scan_id).\
        order_by(severity_sort_logic).all()

    vulnz_list = {}
    for vulnerability in vulnerabilities:
        vuln = {
            'risk_rating': vulnerability.risk_rating.value,
            'cvss_v3_vector': vulnerability.cvss_v3_vector,
            'title': vulnerability.title,
            'short_description': vulnerability.short_description
        }
        vulnz_list[str(vulnerability.id)] = vuln
    if output_format=='json':
        dumper = dumpers.VulnzJsonDumper(output, vulnz_list)
    elif output_format=='csv':
        dumper = dumpers.VulnzCsvDumper(output, vulnz_list)

    try:
        dumper.dump()
    except FileNotFoundError as e:
        console.error(f'No such file or directory: {output}')
        raise click.exceptions.Exit(2) from e

    console.success(f'{len(vulnerabilities)} Vulnerabilities saved to  : {output}')



"""Module responsible for preparing an API request."""
import abc
import logging
from typing import Dict, Optional


logger = logging.getLogger(__name__)


class APIRequest(abc.ABC):
    """API request base class. ALL requests should inherit from this class."""

    @property
    @abc.abstractmethod
    def query(self) -> Optional[str]:
        """Query to the GraphQL API."""
        raise NotImplementedError('Missing implementation')

    @property
    @abc.abstractmethod
    def data(self) -> Optional[Dict]:
        """Body of the API request, containing the query & any additional data."""
        raise NotImplementedError('Missing implementation')



"""models contain the database engine logic and all the db models and the related operations"""

import datetime
import enum
import logging

import sqlalchemy
from sqlalchemy import orm
from sqlalchemy.ext import declarative

from ostorlab import configuration_manager as config_manager
from ostorlab.cli import console as cli_console

logger = logging.getLogger(__name__)
console = cli_console.Console()

ENGINE_URL = f'sqlite:///{config_manager.OSTORLAB_PRIVATE_DIR}/db.sqlite'

metadata = sqlalchemy.MetaData()
Base = declarative.declarative_base(metadata=metadata)


class RiskRating(enum.Enum):
    """Enumeration of the risk rating of a scan."""
    HIGH = 'High'
    MEDIUM = 'Medium'
    LOW = 'Low'
    POTENTIALLY = 'Potentially'
    HARDENING = 'Hardening'
    SECURE = 'Secure'
    IMPORTANT = 'Important'
    INFO = 'Info'


class ScanProgress(enum.Enum):
    NOT_STARTED = 'not_started'
    IN_PROGRESS = 'in_progress'
    STOPPED = 'stopped'
    DONE = 'done'
    ERROR = 'error'


class Database:
    """Handles all Database instantiation and calls."""

    def __init__(self):
        """Constructs the database engine."""
        self._db_engine = sqlalchemy.create_engine(ENGINE_URL)
        self._db_session = None

    @property
    def session(self):
        """Session singleton to run queries on the db engine"""
        if self._db_session is None:
            session_maker = orm.sessionmaker(expire_on_commit=False)
            session_maker.configure(bind=self._db_engine)
            self._db_session = session_maker()
            return self._db_session
        else:
            return self._db_session

    def create_db_tables(self):
        """Create the database tables."""
        with self._db_engine.begin() as conn:
            metadata.create_all(conn)
            logger.info('Tables created')

    def drop_db_tables(self):
        """Drop the database tables."""
        metadata.drop_all(self._db_engine)
        logger.info('Tables dropped')


class Scan(Base):
    """The Scan model"""
    __tablename__ = 'scan'
    id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
    title = sqlalchemy.Column(sqlalchemy.String(255))
    asset = sqlalchemy.Column(sqlalchemy.String(255))
    created_time = sqlalchemy.Column(sqlalchemy.DateTime)
    progress = sqlalchemy.Column(sqlalchemy.Enum(ScanProgress))

    @staticmethod
    def create(asset, title: str = ''):
        """Persist the scan in the database.

        Args:
            title: Scan title.
        Returns:
            Scan object.
        """
        scan = Scan(title=title, asset=asset, created_time=datetime.datetime.now(), progress='NOT_STARTED')
        database = Database()
        database.session.add(scan)
        database.session.commit()
        return scan


class Vulnerability(Base):
    """The Vulnerability model"""
    __tablename__ = 'vulnerability'
    id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
    technical_detail = sqlalchemy.Column(sqlalchemy.Text)
    risk_rating = sqlalchemy.Column(sqlalchemy.Enum(RiskRating))
    cvss_v3_vector = sqlalchemy.Column(sqlalchemy.String(1024))
    dna = sqlalchemy.Column(sqlalchemy.String(256))
    title = sqlalchemy.Column(sqlalchemy.String(256))
    short_description = sqlalchemy.Column(sqlalchemy.Text)
    description = sqlalchemy.Column(sqlalchemy.Text)
    recommendation = sqlalchemy.Column(sqlalchemy.Text)
    scan_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey('scan.id'))

    @staticmethod
    def create(scan_id: int, title: str, short_description: str, description: str,
               recommendation: str, technical_detail: str, risk_rating: str,
               cvss_v3_vector: str, dna: str):
        """Persist the vulnerability in the database.

        Args:
            scan_id: Scan id of the vulnerability.
            title: Vulnerability title.
            false_positive: Boolean to True if the vulnerability is a false positive.
            dna: A unique string to determine a vulnerability and avoid duplicates in the database.
            cvss_v3_vector: Common vulnerability scoring system value.
            risk_rating: Vulnerability risk rating.
            technical_detail: Technical details to demonstrate the finding.
            description: A generic description of the vulnerability.
            short_description: A short description of the vulnerability.
            recommendation: How to address or avoid the vulnerability
        Returns:
            Vulnerability object.
        """
        vuln = Vulnerability(scan_id=scan_id, title=title, short_description=short_description,
                             description=description, recommendation=recommendation,
                             technical_detail=technical_detail, risk_rating=risk_rating, cvss_v3_vector=cvss_v3_vector,
                             dna=dna)
        database = Database()
        database.session.add(vuln)
        database.session.commit()
        return vuln


class ScanStatus(Base):
    """The Scan model"""
    __tablename__ = 'scan_status'
    id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
    created_time = sqlalchemy.Column(sqlalchemy.DateTime)
    key = sqlalchemy.Column(sqlalchemy.String(255))
    value = sqlalchemy.Column(sqlalchemy.Text)
    scan_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey('scan.id'))

    @staticmethod
    def create(key: str, value: str, scan_id: int):
        """Persist the scan status in the database.

        Args:
            key: Scan status key.
            value: Scan status value.
        Returns:
            Scan status object.
        """
        scan_status = ScanStatus(key=key, created_time=datetime.datetime.now(), value=value, scan_id=scan_id)
        database = Database()
        database.session.add(scan_status)
        database.session.commit()
        return scan_status



"""Module responsible for dumping data to different formats."""

import abc
import json
import csv
from typing import Dict


FIELDNAMES = ['id', 'title', 'risk_rating', 'cvss_v3_vector', 'short_description']


class VulnzDumper(abc.ABC):
    """Dumper Base class: All dumpers should inherit from this class to access the dump method."""

    def __init__(self, output_path: str, data: Dict[str, Dict]) -> None:
        """Constructs all the necessary attributes for the object.

        Args:
            output_path: path to the output file.
            data: dictionary, with vulnerability id as a key,
                  and a dictionary of the vulnerability information as values.

        Returns:
            None
        """
        self.output_path = output_path
        self.data = data

    @abc.abstractmethod
    def dump(self) -> None:
        """Dump the vulnerabilities in the right format."""
        raise NotImplementedError('Missing implementation')


class VulnzJsonDumper(VulnzDumper):
    """Vulnerability dumper to json."""

    def dump(self) -> None:
        """Dump vulnerabilities to json file.

        Raises:
            FileNotFoundError: in case the path or file name are invalid.
        """
        if not self.output_path.endswith('.json'):
            self.output_path+= '.json'
        with open(self.output_path , 'w', encoding='utf-8') as outfile:
            json.dump(self.data, outfile)


class VulnzCsvDumper(VulnzDumper):
    """Vulnerability dumper to csv."""

    def dump(self) -> None:
        """Dump vulnerabilities to csv file.

        Raises:
            FileNotFoundError: in case the path or file name are invalid.
        """
        if not self.output_path.endswith('.csv'):
            self.output_path+= '.csv'
        with open(self.output_path , 'w', encoding='utf-8') as outfile:
            csv_writer = csv.DictWriter(outfile, fieldnames = FIELDNAMES)
            csv_writer.writeheader()
            for key in self.data:
                csv_writer.writerow({field: self.data[key].get(field) or key for field in FIELDNAMES})
        


"""Handles any configurations related to Ostorlab, such as storing and retrieving tokens or API keys."""

import json
import pathlib
from datetime import datetime
from typing import Dict, Optional

OSTORLAB_PRIVATE_DIR = pathlib.Path.home() / '.ostorlab'


class ConfigurationManager:
    """Handles any configurations related to Ostorlab, such as storing and retrieving
       API keys.
    """

    def __init__(self, private_dir: pathlib.Path = OSTORLAB_PRIVATE_DIR):
        """Constructs all the necessary attributes for the object
        Args:
            private_dir: The private directory where Ostorlab configurations are stored.
            Defaults to OSTORLAB_PRIVATE_DIR.
        """
        self._private_dir = private_dir
        self._private_dir.mkdir(parents=True, exist_ok=True)
        self._complete_api_key_path = self._private_dir / 'key'


    @property
    def conf_path(self) -> pathlib.Path:
        return self._private_dir.resolve()

    def set_api_data(self, secret_key: str, api_key_id: str, expiry_date: Optional[datetime]) -> None:
        """Persists the API data (key, id, expiry date) to a file in the given path.

        Args:
            api_key: The authenticated user's generated API key.
        """

        api_data = {
            'secret_key': secret_key,
            'api_key_id': api_key_id,
            'expiry_date': expiry_date
        }

        with open(self._complete_api_key_path, 'w', encoding='utf-8') as file:
            data = json.dumps(api_data, indent=4)
            file.write(data)

    def get_api_data(self) -> Optional[Dict]:
        """Gets the API data from the location in which it is saved.

        Returns:
            The user's API data if it exists, otherwise returns None.
        """
        try:
            with open(self._complete_api_key_path, 'r', encoding='utf-8') as file:
                return json.loads(file.read())
        except FileNotFoundError:
            return None

    def get_api_key(self) -> Optional[str]:
        """Gets the API key from the location in which it is saved.

        Returns:
            The user's API key if it exists, otherwise returns None.
        """
        api_data = self.get_api_data()
        if api_data is not None:
            return api_data.get('secret_key')
        else:
            return None

    def get_api_key_id(self) -> Optional[str]:
        """Gets the API key id from the location in which it is saved.

        Returns:
            The user's API key id if it exists, otherwise returns None.
        """
        api_data = self.get_api_data()
        if api_data is not None:
            return api_data.get('api_key_id')
        else:
            return None

    def get_api_key_expiry_date(self) -> Optional[datetime]:
        """Gets the API key expiry date from the location in which it is saved.

        Returns:
            The user's API key expiry date if it exists, otherwise returns None.
        """
        api_data = self.get_api_data()
        if api_data is not None:
            return api_data.get('expiry_date')
        else:
            return None

    def delete_api_data(self) -> None:
        """Deletes the file containing the API data."""
        self._complete_api_key_path.unlink(missing_ok=True)

    @property
    def is_authenticated(self):
        return self.get_api_key_id() is not None



"""TODO(mohsine):Wrtite docstring."""
import enum
import json
from typing import Dict, Optional, BinaryIO

from . import request


class MobileAssetType(enum.Enum):
    ANDROID = enum.auto()
    IOS = enum.auto()


class Plan(enum.Enum):
    FREE = enum.auto()


class CreateMobileScanAPIRequest(request.APIRequest):
    """TODO(mohsine):Wrtite docstring."""

    def __init__(self, title: str, asset_type: MobileAssetType, plan: Plan, application: BinaryIO):
        self._title = title
        self._asset_type = asset_type
        self._plan = plan
        self._application = application

    @property
    def query(self) -> Optional[str]:
        return """
mutation MobileScan($title: String!, $assetType: String!, $application: Upload!, $plan: String!) {
      createMobileScan(title: $title, assetType:$assetType, application: $application, plan: $plan) {
        scan {
            id
        }
      }
    }
    """

    @property
    def data(self) -> Optional[Dict]:
        data = {
            'operations': json.dumps({'query': self.query,
                                      'variables': {'title': self._title,
                                                    'assetType': self._asset_type.name.lower(),
                                                    'application': None,
                                                    'plan': self._plan.name.lower(),
                                                    }
                                      }
                                     ),
            '0': self._application,
            'map': json.dumps({
                '0': ['variables.application'],
            })
        }
        return data



"""Local runtime runs agents locally.

The local runtime requires Docker Swarm to run robust long-running services with a set of configured services, like
a local RabbitMQ.
"""
import logging
import socket
from typing import List
from typing import Optional

import docker
import requests
import tenacity
from docker.models import services as docker_models_services

from ostorlab import exceptions
from ostorlab.assets import asset as base_asset
from ostorlab.cli import console as cli_console
from ostorlab.cli import docker_requirements_checker
from ostorlab.cli import install_agent
from ostorlab.runtimes import definitions
from ostorlab.runtimes import runtime
from ostorlab.runtimes.local.models import models
from ostorlab.runtimes.local.services import mq
from ostorlab.runtimes.local import agent_runtime


NETWORK = 'ostorlab_local_network'
HEALTHCHECK_HOST = '0.0.0.0'
HEALTHCHECK_PORT = 5000
SECOND = 1000000000
HEALTHCHECK_RETRIES = 5
HEALTHCHECK_TIMEOUT = 1 * SECOND
HEALTHCHECK_START_PERIOD = 2 * SECOND
HEALTHCHECK_INTERVAL = int(SECOND / 2)

logger = logging.getLogger(__name__)
console = cli_console.Console()

ASSET_INJECTION_AGENT_DEFAULT = 'agent/ostorlab/inject_asset'
TRACKER_AGENT_DEFAULT = 'agent/ostorlab/tracker'
LOCAL_PERSIST_VULNZ_AGENT_DEFAULT = 'agent/ostorlab/local_persist_vulnz'

DEFAULT_AGENTS = [
    ASSET_INJECTION_AGENT_DEFAULT,
    TRACKER_AGENT_DEFAULT,
    LOCAL_PERSIST_VULNZ_AGENT_DEFAULT
]


class UnhealthyService(exceptions.OstorlabError):
    """A service by the runtime is considered unhealthy."""


class DockerNotInstalledError(exceptions.OstorlabError):
    """Docker is not installed."""


class DockerPermissionsDeniedError(exceptions.OstorlabError):
    """User does not have permissions to run docker."""

class AgentNotInstalled(exceptions.OstorlabError):
    """Agent image not installed."""


def _has_container_image(agent: definitions.AgentSettings):
    """Check if container image is available"""
    return agent.container_image is not None


def _is_agent_status_ok(ip: str) -> bool:
    """Agent are expected to expose a healthcheck service on port 5000 that returns status code 200 and `OK` response.

    Args:
        ip: The agent IP address.

    Returns:
        True if the agent is healthy, false otherwise.
    """
    status_ok = False
    try:
        status_ok = requests.get(f'http://{ip}:5000/status').text == 'OK'
    except requests.exceptions.ConnectionError:
        logger.error('unable to connect to %s', ip)
    return status_ok


def _get_task_ips(service: docker_models_services.Service) -> List[str]:
    """Returns list of IP addresses assigned to the tasks of a docker service.

    Args:
        service: docker service.

    Returns:
        List of IP addresses.
    """
    # current implementation supports only one task per service.
    logger.info('getting ips for task %s', service.name)
    try:
        ips = socket.gethostbyname_ex(f'tasks.{service.name}')
        logger.info('found ips %s for task %s', ips, service.name)
        return ips[2]
    except socket.gaierror:
        return []


def _is_service_type_run(service: docker_models_services.Service) -> bool:
    """Checks if the service should run once or should be continuously running.

    Args:
        service: Docker service.

    Returns:
        Bool indicating if the service is run-once or long-running.
    """
    return service.attrs['Spec']['TaskTemplate']['RestartPolicy']['Condition'] == 'none'


class LocalRuntime(runtime.Runtime):
    """Local runtime runes agents locally using Docker Swarm.
    Local runtime starts a Vanilla RabbitMQ service, starts all the agents listed in the `AgentRunDefinition`, checks
    their status and then inject the target asset.
    """

    def __init__(self, network: str = NETWORK) -> None:
        super().__init__()
        self._network = network
        self._mq_service: Optional[mq.LocalRabbitMQ] = None
        # TODO(alaeddine): inject docker client to support more complex use-cases.

        if not docker_requirements_checker.is_docker_installed():
            raise DockerNotInstalledError('Docker is not installed.')
        elif not docker_requirements_checker.is_user_permitted():
            raise DockerPermissionsDeniedError('User does not have permissions to run docker.')
        else:
            if not docker_requirements_checker.is_swarm_initialized():
                docker_requirements_checker.init_swarm()
            self._docker_client = docker.from_env()

        self._scan_db: Optional[models.Scan] = None

    @property
    def name(self) -> str:
        """Local runtime instance name."""
        if self._scan_db is not None:
            return str(self._scan_db.id)
        else:
            raise ValueError('Scan not created yet')

    def can_run(self, agent_group_definition: definitions.AgentGroupDefinition) -> bool:
        """Checks if the runtime can run the provided agent run definition.

        Args:
            agent_group_definition: Agent and Agent group definition.

        Returns:
            Always true for the moment as the local runtime doesn't have restrictions on what it can run.
        """
        del agent_group_definition
        return True

    def scan(self, title: str, agent_group_definition: definitions.AgentGroupDefinition,
             asset: base_asset.Asset) -> None:
        """Start scan on asset using the provided agent run definition.

        The scan takes care of starting all the scan required services, ensuring they are healthy, starting all the
         agents, ensuring they are healthy and then injects the target asset.

        Args:
            title: Scan title
            agent_group_definition: Agent run definition defines the set of agents and how agents are configured.
            asset: the target asset to scan.

        Returns:
            None
        """
        try:
            console.info('Creating scan entry')
            self._scan_db = self._create_scan_db(asset=str(asset), title=title)
            console.info('Creating network')
            self._create_network()
            console.info('Starting services')
            self._start_services()
            console.info('Checking services are healthy')
            self._check_services_healthy()
            console.info('Starting agents')
            self._start_agents(agent_group_definition)
            console.info('Checking agents are healthy')
            self._check_agents_healthy(agent_group_definition)
            console.info('Injecting asset')
            self._inject_asset(asset)
            console.info('Updating scan status')
            self._update_scan_status()
            console.success('Scan created successfully')
        except AgentNotInstalled as e:
            console.error(f'Agent {e} not installed')

    def stop(self, scan_id: str) -> None:
        """Remove a service (scan) belonging to universe with scan_id(Universe Id).

        Args:
            scan_id: The id of the scan to stop.
        """

        stopped_services = []
        stopped_network = []
        stopped_configs = []
        client = docker.from_env()
        services = client.services.list()
        for service in services:
            service_labels = service.attrs['Spec']['Labels']
            logger.debug('comparing %s and %s', service_labels.get('ostorlab.universe'), scan_id)
            if service_labels.get('ostorlab.universe') == scan_id:
                stopped_services.append(service)
                service.remove()

        networks = client.networks.list()
        for network in networks:
            network_labels = network.attrs['Labels']
            if network_labels is not None and network_labels.get('ostorlab.universe') == scan_id:
                stopped_network.append(network)
                network.remove()

        configs = client.configs.list()
        for config in configs:
            config_labels = config.attrs['Spec']['Labels']
            if config_labels.get('ostorlab.universe') == scan_id:
                stopped_configs.append(config)
                config.remove()

        if stopped_services and stopped_network and stopped_configs:
            console.success('All scan services stopped.')
            database = models.Database()
            session = database.session
            scan = session.query(models.Scan).get(int(scan_id))
            scan.progress = 'STOPPED'
            session.commit()
            console.success('Scan stopped successfully.')
        else:
            console.error(f'Scan with id {scan_id} not found.')

    def _create_scan_db(self, title: str, asset: str):
        """Persist the scan in the database"""
        models.Database().create_db_tables()
        return models.Scan.create(title=title, asset=asset)

    def _update_scan_status(self):
        """Update scan status to in progress"""
        database = models.Database()
        session = database.session
        scan = session.query(models.Scan).get(self._scan_db.id)
        scan.progress = 'IN_PROGRESS'
        session.commit()

    def _create_network(self):
        """Creates a docker swarm network where all services and agents can communicates."""
        if any(network.name == self._network for network in self._docker_client.networks.list()):
            logger.warning('network already exists.')
        else:
            logger.info('creating private network %s', self._network)
            return self._docker_client.networks.create(
                name=self._network,
                driver='overlay',
                attachable=True,
                labels={'ostorlab.universe': self.name},
                check_duplicate=True
            )

    def _start_services(self):
        """Start all the local runtime services."""
        self._start_mq_service()

    def _start_mq_service(self):
        """Start a local rabbitmq service."""
        self._mq_service = mq.LocalRabbitMQ(name=self.name, network=self._network)
        self._mq_service.start()

    def _check_services_healthy(self):
        """Check if the rabbitMQ service is running and healthy."""
        if self._mq_service is None or self._mq_service.is_healthy is False:
            raise UnhealthyService('MQ service is unhealthy.')

    def _check_agents_healthy(self, agent_group_definition: definitions.AgentGroupDefinition):
        """Checks if an agent is healthy."""
        return self._are_agents_ready(agent_group_definition.agents)

    def _start_agents(self, agent_group_definition: definitions.AgentGroupDefinition):
        """Starts all the agents as list in the agent run definition."""
        for agent in agent_group_definition.agents:
            self._start_agent(agent, extra_configs=[])
        self._start_persist_vulnz_agent()
        self._start_tracker_agent()

    def _start_agent(self, agent: definitions.AgentSettings,
                     extra_configs: Optional[List[docker.types.ConfigReference]] = None) -> None:
        """Start agent based on provided definition.

        Args:
            agent: An agent definition containing all the settings of how agent should run and what arguments to pass.
        """
        logger.info('starting agent %s with %s', agent.key, agent.args)

        if _has_container_image(agent) is False:
            raise AgentNotInstalled(agent.key)

        runtime_agent = agent_runtime.AgentRuntime(agent, self.name, self._docker_client, self._mq_service)
        agent_service = runtime_agent.create_agent_service(self._network, extra_configs)

        if agent.replicas > 1:
            # Ensure the agent service had to
            # TODO(alaeddine): Check if sleep if really needed and if it is, implement a parallel way to start agents
            #  and scale them.
            # time.sleep(10)
            self._scale_service(agent_service, agent.replicas)

    @tenacity.retry(stop=tenacity.stop_after_attempt(20),
                    wait=tenacity.wait_exponential(multiplier=1, max=12),
                    # return last value and don't raise RetryError exception.
                    retry_error_callback=lambda lv: lv.outcome.result(),
                    retry=tenacity.retry_if_result(lambda v: v is False))
    def _is_service_healthy(self, service: docker_models_services.Service, replicas=None) -> bool:
        """Checks if a docker service is healthy by checking all tasks status."""
        logger.debug('checking Spec service %s', service.name)
        if not replicas:
            replicas = service.attrs['Spec']['Mode']['Replicated']['Replicas']
        return replicas == len([task for task in service.tasks() if task['Status']['State'] == 'running'])

    def _list_agent_services(self):
        """List the services of type agents. All agent service must start with agent_."""
        services = self._docker_client.services.list(filters={'label': f'ostorlab.universe={self.name}'})
        for service in services:
            if service.name.startswith('agent_'):
                yield service

    def _start_tracker_agent(self):
        """Start the tracker agent to handle the scan lifecycle."""
        tracker_agent_settings = definitions.AgentSettings(key=TRACKER_AGENT_DEFAULT)
        self._start_agent(agent=tracker_agent_settings, extra_configs=[])

    def _start_persist_vulnz_agent(self):
        """Start the local persistence agent to dump vulnerabilities in the local config."""
        persist_vulnz_agent_settings = definitions.AgentSettings(
            key=LOCAL_PERSIST_VULNZ_AGENT_DEFAULT,
            mounts=[])
        self._start_agent(agent=persist_vulnz_agent_settings, extra_configs=[])

    def _inject_asset(self, asset: base_asset.Asset):
        """Injects the scan target assets."""
        asset_config = self._docker_client.configs.create(name='asset', labels={'ostorlab.universe': self.name},
                                                          data=asset.to_proto())
        asset_config_reference = docker.types.ConfigReference(config_id=asset_config.id,
                                                              config_name='asset',
                                                              filename='/tmp/asset.binproto')
        selector_config = self._docker_client.configs.create(name='asset_selector',
                                                             labels={'ostorlab.universe': self.name},
                                                             data=asset.selector)
        selector_config_reference = docker.types.ConfigReference(config_id=selector_config.id,
                                                                 config_name='asset_selector',
                                                                 filename='/tmp/asset_selector.txt')
        inject_asset_agent_settings = definitions.AgentSettings(key=ASSET_INJECTION_AGENT_DEFAULT,
                                                                restart_policy='none')
        self._start_agent(agent=inject_asset_agent_settings,
                          extra_configs=[asset_config_reference, selector_config_reference])

    def _scale_service(self, service: docker_models_services.Service, replicas: int) -> None:
        """Calling scale directly on the service causes an API error. This is a workaround that simulates refreshing
         the service object, then calling the scale API."""
        for s in self._docker_client.services.list():
            if s.name == service.name:
                s.scale(replicas)

    def list(self, page: int = 1, number_elements: int = 10) -> List[runtime.Scan]:
        """Lists scans managed by runtime.

        Args:
            page: Page number for list pagination (default 1).
            number_elements: count of elements to show in the listed page (default 10).

        Returns:
            List of scan objects.
        """
        if page is not None:
            console.warning('Local runtime ignores scan list pagination')

        scans = {}
        database = models.Database()
        database.create_db_tables()
        session = database.session
        for s in session.query(models.Scan):
            scans[s.id] = runtime.Scan(
                id=s.id,
                asset=s.asset,
                created_time=s.created_time,
                progress=s.progress.value,
            )

        universe_ids = set()
        client = docker.from_env()
        services = client.services.list()

        for s in services:
            try:
                service_labels = s.attrs['Spec']['Labels']
                ostorlab_universe_id = service_labels.get('ostorlab.universe')
                if 'ostorlab.universe' in service_labels.keys() and ostorlab_universe_id not in universe_ids:
                    universe_ids.add(ostorlab_universe_id)
                    if ostorlab_universe_id.isnumeric() and int(ostorlab_universe_id) not in scans:
                        console.warning(f'Scan {ostorlab_universe_id} has not traced in DB.')
            except KeyError:
                logger.warning('The label ostorlab.universe do not exist.')

        return list(scans.values())

    @tenacity.retry(stop=tenacity.stop_after_attempt(20),
                    wait=tenacity.wait_exponential(multiplier=1, max=20),
                    # return last value and don't raise RetryError exception.
                    retry_error_callback=lambda lv: lv.result(),
                    retry=tenacity.retry_if_result(lambda v: v is False))
    def _are_agents_ready(self, agents: List[definitions.AgentSettings], fail_fast=True) -> bool:
        """Checks that all agents are ready and healthy while taking into account the run type of agent
         (once vs long-running)."""
        logger.info('listing services ...')
        agent_services = list(self._list_agent_services())
        if len(agent_services) < len(agents):
            logger.error('found %d, expecting %d', len(agent_services), len(agents))
            return False
        else:
            logger.info('found correct count of services')

        for service in agent_services:
            logger.info('checking %s ...', service.name)
            if not _is_service_type_run(service):
                if self._is_service_healthy(service):
                    logger.info('agent service %s is healthy', service.name)
                else:
                    logger.error('agent service %s is not healthy', service.name)
                    if fail_fast:
                        return False
        return True

    def install(self) -> None:
        """Installs the default agents.

        Returns:
            None
        """
        for agent_key in DEFAULT_AGENTS:
            install_agent.install(agent_key=agent_key)

